@@ -19,14 +19,17 @@ use std::{
 use libseccomp::ScmpSyscall;
 use nix::{
     errno::Errno,
-    fcntl::{AtFlags, OFlag, OpenHow},
-    sys::socket::{bind, connect, AddressFamily, Shutdown, SockFlag, SockType, SockaddrLike},
+    fcntl::{AtFlags, OFlag},
+    sys::{
+        socket::{bind, connect, AddressFamily, Shutdown, SockFlag, SockType, SockaddrLike},
+        stat::Mode,
+    },
     unistd::UnlinkatFlags,
     NixPath,
 };
 
 use crate::{
-    compat::{MFdFlags, RenameFlags},
+    compat::{MFdFlags, OpenHow, RenameFlags},
     rng::randint,
     uts::UtsName,
 };
@@ -63,6 +66,21 @@ pub(crate) static OPENAT2_COOKIE_ARG4: LazyLock<Cookie> =
 #[expect(clippy::disallowed_methods)]
 pub(crate) static OPENAT2_COOKIE_ARG5: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static FCHDIR_COOKIE_ARG1: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static FCHDIR_COOKIE_ARG2: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static FCHDIR_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static FCHDIR_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static FCHDIR_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 /// System call argument cookie: socket(2) $arg3
 #[expect(clippy::disallowed_methods)]
 pub static SOCKET_COOKIE_ARG3: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
@@ -184,6 +202,21 @@ pub(crate) static FTRUNCATE64_COOKIE_ARG4: LazyLock<Cookie> =
 pub(crate) static FTRUNCATE64_COOKIE_ARG5: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
 #[expect(clippy::disallowed_methods)]
+pub(crate) static UMASK_COOKIE_ARG1: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static UMASK_COOKIE_ARG2: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static UMASK_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static UMASK_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static UMASK_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
 pub(crate) static UNAME_COOKIE_ARG1: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
 #[expect(clippy::disallowed_methods)]
@@ -271,7 +304,7 @@ pub(crate) fn safe_openat2<Fd: AsFd, P: NixPath + ?Sized>(
             dirfd.as_fd().as_raw_fd(),
             cstr.as_ptr(),
             std::ptr::addr_of_mut!(how),
-            size_of::<libc::open_how>(),
+            size_of::<OpenHow>(),
             *OPENAT2_COOKIE_ARG4,
             *OPENAT2_COOKIE_ARG5,
         )
@@ -939,6 +972,24 @@ pub(crate) fn safe_linkat<Fd1: AsFd, Fd2: AsFd, P1: NixPath + ?Sized, P2: NixPat
     Errno::result(res).map(drop)
 }
 
+/// Safe fchdir(2) confined by syscall cookies.
+#[inline(always)]
+pub fn safe_fchdir<Fd: AsFd>(dirfd: Fd) -> Result<(), Errno> {
+    // SAFETY: In libc we trust.
+    Errno::result(unsafe {
+        libc::syscall(
+            libc::SYS_fchdir,
+            dirfd.as_fd().as_raw_fd(),
+            *FCHDIR_COOKIE_ARG1,
+            *FCHDIR_COOKIE_ARG2,
+            *FCHDIR_COOKIE_ARG3,
+            *FCHDIR_COOKIE_ARG4,
+            *FCHDIR_COOKIE_ARG5,
+        )
+    })
+    .map(drop)
+}
+
 /// Safe uname(2) confined by syscall cookies.
 #[inline(always)]
 pub fn safe_uname() -> Result<UtsName, Errno> {
@@ -960,6 +1011,27 @@ pub fn safe_uname() -> Result<UtsName, Errno> {
     Ok(name)
 }
 
+/// Safe (2) confined by syscall cookies.
+#[inline(always)]
+pub fn safe_umask(mode: Mode) -> Mode {
+    // SAFETY: In libc we trust.
+    #[expect(clippy::cast_possible_truncation)]
+    #[expect(clippy::cast_sign_loss)]
+    let prev = unsafe {
+        libc::syscall(
+            libc::SYS_umask,
+            mode.bits(),
+            *UMASK_COOKIE_ARG1,
+            *UMASK_COOKIE_ARG2,
+            *UMASK_COOKIE_ARG3,
+            *UMASK_COOKIE_ARG4,
+            *UMASK_COOKIE_ARG5,
+        )
+    } as libc::mode_t;
+    #[expect(clippy::disallowed_methods)]
+    Mode::from_bits(prev).expect("[BUG] umask returned invalid Mode")
+}
+
 /// Safe pipe2(2) confined by syscall cookies.
 #[inline(always)]
 pub fn safe_pipe2(flags: OFlag) -> Result<(OwnedFd, OwnedFd), Errno> {
