@@ -150,7 +150,7 @@ use nix::{
 };
 
 use crate::{
-    fs::{fd_status_flags, to_valid_fd},
+    fd::{fd_status_flags, to_valid_fd},
     log::log_is_main,
     log_enabled,
     lookup::{CanonicalPath, FileInfo, FileType},
@@ -626,7 +626,7 @@ pub(crate) fn to_sflag(arg: u64) -> Result<SFlag, Errno> {
 #[cfg(test)]
 mod tests {
     use super::*;
-    use crate::fs::AT_EXECVE_CHECK;
+    use crate::fd::AT_EXECVE_CHECK;
 
     #[test]
     fn test_to_atflags() {
