@@ -21,23 +21,24 @@ use std::{
 use libseccomp::ScmpNotifResp;
 use nix::{
     errno::Errno,
-    fcntl::{OFlag, OpenHow, ResolveFlag},
-    sys::stat::{umask, Mode},
-    unistd::Uid,
+    fcntl::OFlag,
+    sys::stat::Mode,
+    unistd::{Pid, Uid},
     NixPath,
 };
 
 use crate::{
-    compat::{fstatx, STATX_MODE, STATX_UID},
+    compat::{fstatx, FsType, OpenHow, ResolveFlag, STATX_MODE, STATX_UID},
     config::{
         ENV_ID, ESYD_SH, MAGIC_PREFIX, NULL_FD, PINK_FLOYD, PROC_FILE, SAFE_MFD_FLAGS,
         SEE_EMILY_PLAY, SYD_EL, THE_PIPER,
     },
     confine::EIDRM,
-    cookie::{safe_memfd_create, safe_openat2, safe_uname},
+    cookie::{safe_memfd_create, safe_openat2, safe_umask, safe_uname},
     err::err2no,
     error,
-    fs::{format_oflags, is_dev_kfd, oflag_accmode, oflag_nonblock, seal_memfd},
+    fd::{is_dev_kfd, seal_memfd_all},
+    fs::{format_oflags, oflag_accmode, oflag_nonblock},
     hash::aes_ctr_tmp,
     info, log_enabled,
     lookup::{
@@ -47,7 +48,7 @@ use crate::{
     ofd::{lock_fd, unlock_fd},
     path::{XPath, XPathBuf},
     proc::{proc_tty, proc_umask},
-    req::{SysArg, UNotifyEventRequest},
+    req::{SysArg, SysFlags, UNotifyEventRequest},
     retry::retry_on_eintr,
     sandbox::{Action, Capability, Sandbox},
     syslog::LogLevel,
@@ -259,7 +260,7 @@ fn syscall_open_handler(
             return Err(Errno::EACCES);
         }
 
-        if sandbox.verbose {
+        if sandbox.log_scmp() {
             crate::debug!("ctx": "open", "op": "init",
                 "cap": caps,
                 "sbc": *sandbox.state,
@@ -291,9 +292,10 @@ fn syscall_open_handler(
         }
 
         // Read the remote path and check for magic path.
-        let (mut path, magic, _) = request.read_path(&sandbox, arg, true)?;
+        arg.flags.insert(SysFlags::CHECK_MAGIC);
+        let (mut path, magic, _) = request.read_path(&sandbox, arg)?;
 
-        if sandbox.verbose {
+        if sandbox.log_scmp() {
             crate::debug!("ctx": "open", "op": "read_path",
                 "path": &path, "magic": magic,
                 "req": &request);
@@ -346,7 +348,7 @@ fn syscall_open_handler(
                 file.rewind().or(Err(Errno::EIO))?;
 
                 // SAFETY: Deny further writes to the file descriptor.
-                seal_memfd(&file)?;
+                seal_memfd_all(&file)?;
 
                 MaybeFd::Owned(file.into())
             } else {
@@ -500,23 +502,10 @@ fn syscall_open_handler(
         let mut action = Action::Allow;
         let mut filter = false;
 
-        let mut path_check = if !caps.is_empty() {
-            // Convert /proc/${pid} to /proc/self as necessary.
-            let path_check = if let Some(p) = path.abs().split_prefix(b"/proc") {
-                let mut buf = itoa::Buffer::new();
-                let req = request.scmpreq;
-                let pid = buf.format(req.pid);
-                if let Some(p) = p.split_prefix(pid.as_bytes()) {
-                    let mut pdir = XPathBuf::from("/proc/self");
-                    pdir.push(p.as_bytes());
-                    Cow::Owned(pdir)
-                } else {
-                    Cow::Borrowed(path.abs())
-                }
-            } else {
-                Cow::Borrowed(path.abs())
-            };
+        // Convert /proc/$pid to /proc/self as necessary.
+        let mut path_check = path.abs().replace_proc_self(pid);
 
+        if !caps.is_empty() {
             // If open is a TTY both the virtual and real path are checked for access.
             // The more critical action wins.
             let mut flip_tty = false;
@@ -538,13 +527,9 @@ fn syscall_open_handler(
             }
             #[expect(clippy::disallowed_methods)]
             if flip_tty {
-                path_tty.unwrap()
-            } else {
-                path_check
+                path_check = path_tty.unwrap()
             }
-        } else {
-            Cow::Borrowed(path.abs())
-        };
+        }
 
         let (hidden, mut crypted) = if action.is_denying() {
             // No need to check for mask when denying.
@@ -589,7 +574,7 @@ fn syscall_open_handler(
                 flags.insert(OFlag::O_APPEND);
                 flags.remove(OFlag::O_TRUNC);
 
-                if sandbox.verbose {
+                if sandbox.log_scmp() {
                     info!("ctx": "open", "op": "set_append_only",
                             "msg": "added O_APPEND and removed O_TRUNC from open flags",
                             "sys": request.syscall, "path": &path_check,
@@ -609,8 +594,50 @@ fn syscall_open_handler(
             (false, crypted)
         };
 
+        // SAFETY:
+        // Check for Filesystem sandboxing as necessary.
+        let log_scmp = sandbox.log_scmp();
+        let mut maybe_fs_type = None;
+        if action.is_allowing() && sandbox.has_fs() {
+            let fd = path.dir.as_ref().ok_or(Errno::EBADF)?;
+            let fs_type = FsType::get(fd)?;
+            let (action, filter) = sandbox.check_fs(fs_type);
+            if !filter && action >= Action::Warn && log_enabled!(LogLevel::Warn) {
+                if log_scmp {
+                    warn!("ctx": "access", "cap": Capability::CAP_FS, "act": action,
+                        "sys": request.syscall, "fs": fs_type, "path": &path_check,
+                        "mode": mode.bits(),
+                        "oflags": format_oflags(flags), "rflags": arg.fsflags,
+                        "type": path.typ,
+                        "tip": format!("configure `allow/fs+{fs_type}'"),
+                        "req": &request);
+                } else {
+                    warn!("ctx": "access", "cap": Capability::CAP_FS, "act": action,
+                        "sys": request.syscall, "fs": fs_type, "path": &path_check,
+                        "mode": mode.bits(),
+                        "oflags": format_oflags(flags), "rflags": arg.fsflags,
+                        "type": path.typ,
+                        "tip": format!("configure `allow/fs+{fs_type}'"),
+                        "pid": request.scmpreq.pid);
+                }
+            }
+
+            match action {
+                Action::Allow | Action::Warn => {}
+                Action::Deny | Action::Filter if hidden => return Err(Errno::ENOENT),
+                Action::Deny | Action::Filter => return Err(Errno::EACCES),
+                Action::Panic => panic!(),
+                Action::Exit => std::process::exit(libc::EACCES),
+                action => {
+                    // Stop|Kill
+                    let _ = request.kill(action);
+                    return Err(Errno::EACCES);
+                }
+            }
+            maybe_fs_type = Some(fs_type);
+        }
+
         let force_umask = sandbox.umask;
-        let verbose = sandbox.verbose;
         let unsafe_open_kfd = sandbox.flags.allow_unsafe_open_kfd();
         let unsafe_open_path = sandbox.flags.allow_unsafe_open_path();
         let filter_machine_id = !sandbox.flags.allow_unsafe_machine_id();
@@ -626,10 +653,18 @@ fn syscall_open_handler(
 
         // Perform action: allow->emulate, deny->log.
         if !filter && action >= Action::Warn && log_enabled!(LogLevel::Warn) {
+            let fs_type = if let Some(fs_type) = maybe_fs_type {
+                fs_type
+            } else {
+                path.dir
+                    .as_ref()
+                    .ok_or(Errno::EBADF)
+                    .and_then(FsType::get)?
+            };
             let grp = caps.to_string().to_ascii_lowercase();
-            if verbose {
+            if log_scmp {
                 warn!("ctx": "access", "cap": caps, "act": action,
-                    "sys": request.syscall, "path": &path_check,
+                    "sys": request.syscall, "fs": fs_type, "path": &path_check,
                     "mode": mode.bits(),
                     "oflags": format_oflags(flags), "rflags": arg.fsflags,
                     "type": path.typ,
@@ -637,7 +672,7 @@ fn syscall_open_handler(
                     "req": &request);
             } else {
                 warn!("ctx": "access", "cap": caps, "act": action,
-                    "sys": request.syscall, "path": &path_check,
+                    "sys": request.syscall, "fs": fs_type, "path": &path_check,
                     "mode": mode.bits(),
                     "oflags": format_oflags(flags), "rflags": arg.fsflags,
                     "type": path.typ,
@@ -711,7 +746,7 @@ fn syscall_open_handler(
         if restrict_create && o_creat && matches!(path.typ, Some(FileType::Reg | FileType::Fifo)) {
             let fd = path.dir.as_ref().map(|fd| fd.as_fd()).ok_or(Errno::EBADF)?;
             // assert fd refers to file.
-            assert!(path.base.is_empty());
+            assert!(path.base().is_empty());
 
             let statx = fstatx(fd, STATX_MODE | STATX_UID)?;
             let mode = libc::mode_t::from(statx.stx_mode);
@@ -732,7 +767,7 @@ fn syscall_open_handler(
             }
 
             if let Some(msg) = err {
-                if verbose {
+                if log_scmp {
                     error!("ctx": "trusted_file_creation", "cap": caps,
                         "sys": request.syscall, "path": &path_check,
                         "mode": mode, "oflags": format_oflags(flags), "rflags": arg.fsflags,
@@ -821,7 +856,7 @@ fn syscall_open_handler(
                 // 4. Mode must be 0 if O_CREAT or O_TMPFILE is not in flags.
                 // Note, magic symbolic links are an exception here.
                 if matches!(path.typ, Some(FileType::MagicLnk(_))) {
-                    if !path.base.is_empty() || flags.contains(OFlag::O_NOFOLLOW) {
+                    if !path.base().is_empty() || flags.contains(OFlag::O_NOFOLLOW) {
                         // Magic symlink O_PATH fds were continued
                         // as necessary if relevant unsafe options
                         // were set. After this point, we have to
@@ -866,7 +901,7 @@ fn syscall_open_handler(
                         // can skip it gracefully.
                         return Ok(ScmpNotifResp::new(0, 0, EIDRM, 0));
                     }
-                } else if path.base.is_empty() && !o_tmpfl {
+                } else if path.base().is_empty() && !o_tmpfl {
                     if let Some(fd) = do_open_pre(&request, &path, filter_machine_id, filter_uname)?
                     {
                         // pre-open hook intercepted, no need to open.
@@ -875,7 +910,7 @@ fn syscall_open_handler(
                         // Existing path: Construct path to proc(5) magic symlink,
                         // or pre-open file descriptor.
                         let fd = do_open(&request, &path, flags, is_blocking)?;
-                        do_open_post(fd, path, filter_proc_status)?
+                        do_open_post(fd, request.scmpreq.pid(), path, filter_proc_status)?
                     }
                 } else if o_mask {
                     // Cannot be O_PATH or encrypted fd!
@@ -923,7 +958,7 @@ fn handle_crypt(
     opts: CryptOpts,
 ) -> Result<Option<OwnedFd>, Errno> {
     #[expect(clippy::disallowed_methods)]
-    let files = request.crypt_map.as_ref().unwrap();
+    let files = request.cache.crypt_map.as_ref().unwrap();
     let my_mode = AesMod::from(safe_flags);
 
     // Check for concurrent read access to the same crypt fd.
@@ -981,10 +1016,10 @@ fn handle_crypt(
 
     // Open or create the file as read-write.
     //
-    // Note, if `path.base.is_empty()` we have an existing file
+    // Note, if `path.base().is_empty()` we have an existing file
     // which do_open_crypt is going to reopen using proc(5) magic
     // symlink.
-    let create = if path.base.is_empty() {
+    let create = if path.base().is_empty() {
         // Strip O_CREAT when NOT creating.
         flags.remove(OFlag::O_CREAT);
         // O_TMPFILE is valid if base is empty, when
@@ -1102,15 +1137,15 @@ fn do_creat(
     let req = request.scmpreq;
     let mask = proc_umask(req.pid())?;
 
-    // `path.base` may be empty for O_TMPFILE.
-    let base = if !path.base.is_empty() {
-        path.base
+    // `path.base()` may be empty for O_TMPFILE.
+    let base = if !path.base().is_empty() {
+        path.base()
     } else {
         XPath::from_bytes(b".")
     };
 
     // Set umask which is per-thread here.
-    umask(mask);
+    safe_umask(mask);
 
     // All set, make the openat2(2) call.
     let fd = path.dir.as_ref().map(|fd| fd.as_fd()).ok_or(Errno::EBADF)?;
@@ -1154,7 +1189,7 @@ fn to_oflag(arg: u64, truncate: bool) -> Result<OFlag, Errno> {
 // using proc(5) magic symlink, otherwise do_creat() is used.
 fn do_open_crypt(path: &CanonicalPath, mut safe_flags: OFlag) -> Result<OwnedFd, Errno> {
     assert!(
-        path.base.is_empty(),
+        path.base().is_empty(),
         "BUG: non-canonical path passed to do_open_crypt, report a bug!"
     );
     assert!(
@@ -1185,7 +1220,7 @@ fn do_open(
     mut flags: OFlag,
     is_blocking: bool,
 ) -> Result<OwnedFd, Errno> {
-    assert!(path.base.is_empty());
+    assert!(path.base().is_empty());
 
     // SAFETY: Note, the path may be borrowed here, ie it can be a fd to the preopen `/`,
     // `/proc`, or `/dev/null` fds. In these cases, we still want to reopen, because the
@@ -1245,12 +1280,13 @@ fn do_open_pre(
 
 fn do_open_post(
     mut fd: OwnedFd,
+    pid: Pid,
     path: CanonicalPath,
     filter_proc_status: bool,
 ) -> Result<OwnedFd, Errno> {
     // SAFETY: Filter security-sensitive proc_pid_status(5) contents.
-    if filter_proc_status && path.abs().is_proc_pid_status() {
-        fd = do_mask_proc_pid_status(fd, path)?;
+    if filter_proc_status && path.abs().is_proc_status() {
+        fd = do_mask_proc_pid_status(fd, pid, path)?;
     }
 
     Ok(fd)
@@ -1274,7 +1310,7 @@ fn do_open_machine_id() -> Result<OwnedFd, Errno> {
 
     // Rewind file offset and seal to deny further writes.
     mfd.rewind().or(Err(Errno::EIO))?;
-    seal_memfd(&mfd)?;
+    seal_memfd_all(&mfd)?;
 
     Ok(mfd.into())
 }
@@ -1302,7 +1338,7 @@ fn do_open_proc_version(request: &UNotifyEventRequest) -> Result<OwnedFd, Errno>
 
     // Rewind file offset and seal to deny further writes.
     mfd.rewind().or(Err(Errno::EIO))?;
-    seal_memfd(&mfd)?;
+    seal_memfd_all(&mfd)?;
 
     Ok(mfd.into())
 }
@@ -1324,14 +1360,20 @@ fn do_open_proc_osrelease() -> Result<OwnedFd, Errno> {
 
     // Rewind file offset and seal to deny further writes.
     mfd.rewind().or(Err(Errno::EIO))?;
-    seal_memfd(&mfd)?;
+    seal_memfd_all(&mfd)?;
 
     Ok(mfd.into())
 }
 
-fn do_mask_proc_pid_status(fd: OwnedFd, path: CanonicalPath) -> Result<OwnedFd, Errno> {
+fn do_mask_proc_pid_status(fd: OwnedFd, pid: Pid, path: CanonicalPath) -> Result<OwnedFd, Errno> {
     // Prepare the name of the memfd for identification.
+    // Undo /proc/self -> /proc/$pid.
     let mut path = path.take();
+    if path.starts_with(b"/proc/self/") {
+        let mut pfd = XPathBuf::from("/proc");
+        pfd.push_pid(pid);
+        path.replace_prefix(b"/proc/self", pfd.as_bytes())?;
+    }
     path.replace_prefix(b"/", b"syd/")?;
 
     // Open a memory fd to be sealed for write.
@@ -1342,7 +1384,7 @@ fn do_mask_proc_pid_status(fd: OwnedFd, path: CanonicalPath) -> Result<OwnedFd,
 
     // Rewind file offset and seal to deny further writes.
     mfd.rewind().or(Err(Errno::EIO))?;
-    seal_memfd(&mfd)?;
+    seal_memfd_all(&mfd)?;
 
     Ok(mfd.into())
 }
