@@ -28,7 +28,7 @@ use std::{
     ffi::{CStr, OsStr, OsString},
     fmt,
     fs::File,
-    io::{BufReader, IoSlice, IoSliceMut, Read, Seek, SeekFrom},
+    io::{BufReader, Cursor, IoSlice, IoSliceMut, Read, Seek, SeekFrom},
     num::NonZeroUsize,
     ops::Range,
     os::{
@@ -40,10 +40,10 @@ use std::{
 
 use bitflags::bitflags;
 use btoi::{btoi, btoi_radix};
-use memchr::{arch::all::is_prefix, memrchr};
+use memchr::{arch::all::is_prefix, memchr, memrchr};
 use nix::{
     errno::Errno,
-    fcntl::{openat2, OFlag, ResolveFlag},
+    fcntl::OFlag,
     libc::pid_t,
     sys::{
         stat::Mode,
@@ -60,17 +60,19 @@ use procfs_core::{
     },
     CryptoTable, FromBufRead,
 };
+use serde::{ser::SerializeMap, Serialize, Serializer};
 
 use crate::{
-    compat::{getdents64, statx, STATX_INO},
+    compat::{getdents64, openat2, statx, ResolveFlag, STATX_INO},
     config::*,
     confine::SydMemoryMap,
     cookie::{PROCMAP_QUERY_COOKIE_ARG3, PROCMAP_QUERY_COOKIE_ARG4, PROCMAP_QUERY_COOKIE_ARG5},
     elf::ElfType,
     err::{err2no, proc_error_to_errno},
-    fs::{is_empty_file, is_open_fd, parse_fd, readlinkat, AT_BADFD},
+    fd::{is_empty_file, is_open_fd, parse_fd, AT_BADFD},
+    fs::readlinkat,
     hash::{SydHashMap, SydHashSet, SydIndexSet},
-    io::{read_all, write_all},
+    io::{read_all, read_buf, write_all},
     lookup::{safe_open, safe_open_how_abs},
     parsers::{
         map_result,
@@ -128,6 +130,41 @@ bitflags! {
     }
 }
 
+impl fmt::Display for ProcmapQueryFlags {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        if self.contains(Self::VMA_READABLE) {
+            f.write_str("r")?;
+        } else {
+            f.write_str("-")?;
+        }
+        if self.contains(Self::VMA_WRITABLE) {
+            f.write_str("w")?;
+        } else {
+            f.write_str("-")?;
+        }
+        if self.contains(Self::VMA_EXECUTABLE) {
+            f.write_str("x")?;
+        } else {
+            f.write_str("-")?;
+        }
+        if self.contains(Self::VMA_SHARED) {
+            f.write_str("s")?;
+        } else {
+            f.write_str("p")?;
+        }
+        Ok(())
+    }
+}
+
+impl Serialize for ProcmapQueryFlags {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        serializer.collect_str(self)
+    }
+}
+
 impl From<MMPermissions> for ProcmapQueryFlags {
     fn from(perms: MMPermissions) -> Self {
         let mut flags = Self::empty();
@@ -209,15 +246,15 @@ pub struct ProcmapQuery {
     /// (out)
     pub vma_offset: u64,
 
-    /// Backing file’s inode number, or zero if no backing file.
+    /// Backing file's inode number, or zero if no backing file.
     /// (out)
     pub inode: u64,
 
-    /// Backing file’s device major number, or zero if no backing file.
+    /// Backing file's device major number, or zero if no backing file.
     /// (out)
     pub dev_major: u32,
 
-    /// Backing file’s device minor number, or zero if no backing file.
+    /// Backing file's device minor number, or zero if no backing file.
     /// (out)
     pub dev_minor: u32,
 
@@ -389,16 +426,33 @@ impl Default for Status {
     }
 }
 
+impl Serialize for Status {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        let mut map = serializer.serialize_map(Some(7))?;
+        map.serialize_entry("umask", &self.umask.bits())?;
+        map.serialize_entry("pid", &self.pid.as_raw())?;
+        map.serialize_entry("sig_pending_thread", &self.sig_pending_thread)?;
+        map.serialize_entry("sig_pending_process", &self.sig_pending_process)?;
+        map.serialize_entry("sig_blocked", &self.sig_blocked)?;
+        map.serialize_entry("sig_ignored", &self.sig_ignored)?;
+        map.serialize_entry("sig_caught", &self.sig_caught)?;
+        map.end()
+    }
+}
+
 /// Safe `OFlags` to use when opening proc(5) files.
 pub const OFLAG_PROC: OFlag =
     OFlag::from_bits_retain(libc::O_RDONLY | libc::O_CLOEXEC | libc::O_NOFOLLOW | libc::O_NOCTTY);
 
 /// Safe `ResolveFlag` to use when opening proc(5) files.
 pub const RFLAG_PROC: ResolveFlag = ResolveFlag::from_bits_retain(
-    libc::RESOLVE_NO_MAGICLINKS
-        | libc::RESOLVE_NO_SYMLINKS
-        | libc::RESOLVE_BENEATH
-        | libc::RESOLVE_NO_XDEV,
+    ResolveFlag::RESOLVE_NO_MAGICLINKS.bits()
+        | ResolveFlag::RESOLVE_NO_SYMLINKS.bits()
+        | ResolveFlag::RESOLVE_BENEATH.bits()
+        | ResolveFlag::RESOLVE_NO_XDEV.bits(),
 );
 
 // Function to map proc(5) errno ENOENT to ESRCH to reduce confusion.
@@ -501,7 +555,6 @@ pub fn proc_statm(pid: Pid) -> Result<Statm, Errno> {
 }
 
 /// Return the status of the given process.
-#[inline]
 pub fn proc_status(pid: Pid) -> Result<Status, Errno> {
     proc_status_read(proc_status_open(pid)?)
 }
@@ -535,7 +588,7 @@ pub fn proc_tgid(tid: Pid) -> Result<Pid, Errno> {
 
     // Read up to 192 bytes.
     let mut data = [0u8; 192];
-    let nread = read_all(fd, &mut data)?;
+    let nread = read_buf(fd, &mut data)?;
 
     map_result(parse_status_tgid(&data[..nread]))
 }
@@ -548,7 +601,7 @@ pub fn proc_umask(pid: Pid) -> Result<Mode, Errno> {
 
     // Read up to 160 bytes.
     let mut data = [0u8; 160];
-    let nread = read_all(fd, &mut data)?;
+    let nread = read_buf(fd, &mut data)?;
 
     map_result(parse_status_umask(&data[..nread]))
 }
@@ -581,13 +634,8 @@ pub fn proc_stack_pointer(pid: Pid) -> Result<u64, Errno> {
     path.push(b"syscall");
 
     // Safely open proc(5) file.
-    let mut file = safe_open_proc(&path)
-        .map(File::from)
-        .map(BufReader::new)
-        .map_err(proc_errno)?;
-
-    let mut data = Vec::with_capacity(128);
-    file.read_to_end(&mut data).or(Err(Errno::EIO))?;
+    let file = safe_open_proc(&path).map_err(proc_errno)?;
+    let data = read_all(file)?;
 
     // Check if the contents equal "running".
     if is_prefix(&data, b"running") {
@@ -625,19 +673,17 @@ pub fn proc_auxv(pid: Pid) -> Result<SydHashMap<u64, u64>, Errno> {
     path.push(b"auxv");
 
     // Safely open proc(5) file.
-    let mut file = safe_open_proc(&path).map(File::from).map_err(proc_errno)?;
+    let file = safe_open_proc(&path).map(File::from).map_err(proc_errno)?;
 
     let mut map = SydHashMap::default();
 
-    let mut buf = Vec::new();
-    let bytes_read = file.read_to_end(&mut buf).map_err(|err| err2no(&err))?;
-    if bytes_read == 0 {
-        // some kernel processes won't have any data for their auxv file
+    let buf = read_all(&file)?;
+    if buf.is_empty() {
+        // Some kernel processes won't have any data for their auxv file.
         return Ok(map);
     }
-    buf.truncate(bytes_read);
-    let mut file = std::io::Cursor::new(buf);
 
+    let mut file = Cursor::new(buf);
     let mut buf = 0usize.to_ne_bytes();
     loop {
         file.read_exact(&mut buf).map_err(|err| err2no(&err))?;
@@ -673,7 +719,7 @@ pub fn proc_comm(pid: Pid) -> Result<XPathBuf, Errno> {
 
     // Read up to 16 characters or until EOF.
     let mut comm = [0u8; 16];
-    let mut nread = read_all(file, &mut comm)?;
+    let mut nread = read_buf(file, &mut comm)?;
 
     // Remove the trailing NUL-byte and return an `XPathBuf`.
     let idx = nread.saturating_sub(1);
@@ -703,7 +749,7 @@ pub fn proc_cmdline(pid: Pid) -> Result<XPathBuf, Errno> {
     // Read up to 256 bytes.
     const LIMIT: usize = 256;
     let mut data = [0u8; LIMIT];
-    let nread = read_all(file, &mut data)?;
+    let nread = read_buf(file, &mut data)?;
 
     let mut data = data.to_vec();
     // Determine if EOF was reached or if we hit the limit
@@ -820,24 +866,7 @@ pub fn proc_mem(pid: Pid) -> Result<u64, Errno> {
     Ok(sum)
 }
 
-/// Returns the memory map of an address by reading _proc_pid_maps_(5).
-///
-/// Returns EFAULT if the address does not fall within any range.
-pub fn proc_mmap(pid: Pid, addr: u64) -> Result<SydMemoryMap, Errno> {
-    let maps = proc_maps(pid)?;
-
-    for mmap in maps {
-        // Check if the address falls within range.
-        if (mmap.0.address.0..mmap.0.address.1).contains(&addr) {
-            return Ok(mmap);
-        }
-    }
-
-    Err(Errno::EFAULT)
-}
-
 /// Return the system pipe max limit.
-#[inline]
 pub fn proc_pipemax() -> Result<libc::c_int, Errno> {
     proc_pipemax_read(proc_pipemax_open()?)
 }
@@ -851,7 +880,7 @@ pub fn proc_pipemax_open() -> Result<OwnedFd, Errno> {
 pub fn proc_pipemax_read<Fd: AsFd>(fd: Fd) -> Result<libc::c_int, Errno> {
     // Read up to 24 bytes.
     let mut data = [0u8; 24];
-    let nread = read_all(fd, &mut data)?;
+    let nread = read_buf(fd, &mut data)?;
 
     btoi::<libc::c_int>(data[..nread].trim_ascii()).or(Err(Errno::EINVAL))
 }
@@ -865,7 +894,7 @@ pub fn proc_fs_file_max() -> Result<u64, Errno> {
 
     // Read up to 24 bytes + 1 bytes for \n.
     let mut data = [0u8; 25];
-    let nread = read_all(fd, &mut data)?;
+    let nread = read_buf(fd, &mut data)?;
 
     btoi::<u64>(data[..nread].trim_ascii()).or(Err(Errno::EINVAL))
 }
@@ -879,7 +908,7 @@ pub fn proc_fs_nr_open() -> Result<u64, Errno> {
 
     // Read up to 24 bytes + 1 bytes for \n.
     let mut data = [0u8; 25];
-    let nread = read_all(fd, &mut data)?;
+    let nread = read_buf(fd, &mut data)?;
 
     btoi::<u64>(data[..nread].trim_ascii()).or(Err(Errno::EINVAL))
 }
@@ -898,7 +927,7 @@ pub fn proc_mmap_min_addr() -> Result<u64, Errno> {
 
     // Read up to 24 bytes + 1 bytes for \n.
     let mut data = [0u8; 25];
-    let nread = read_all(fd, &mut data)?;
+    let nread = read_buf(fd, &mut data)?;
 
     btoi::<u64>(data[..nread].trim_ascii()).or(Err(Errno::EINVAL))
 }
@@ -933,7 +962,7 @@ pub fn proc_kernel_randomize_va_space() -> Result<u8, Errno> {
 
     // Read up to 1 byte + 1 byte for \n.
     let mut data = [0u8; 2];
-    let nread = read_all(fd, &mut data)?;
+    let nread = read_buf(fd, &mut data)?;
 
     btoi::<u8>(data[..nread].trim_ascii()).or(Err(Errno::EINVAL))
 }
@@ -1250,14 +1279,32 @@ pub struct Vma {
 
     // VMA flags
     flags: ProcmapQueryFlags,
+
+    // VMA name
+    name: [u8; PATH_MAX],
 }
 
 impl Vma {
-    fn new(ptr: NonNull<libc::c_void>, len: NonZeroUsize, flags: ProcmapQueryFlags) -> Self {
-        Self { ptr, len, flags }
+    fn new(
+        ptr: NonNull<libc::c_void>,
+        len: NonZeroUsize,
+        flags: ProcmapQueryFlags,
+        name: [u8; PATH_MAX],
+    ) -> Self {
+        Self {
+            ptr,
+            len,
+            flags,
+            name,
+        }
     }
 
     /// Return VMA address.
+    pub fn addr(&self) -> usize {
+        self.ptr.as_ptr() as usize
+    }
+
+    /// Return VMA pointer.
     pub fn as_ptr(&self) -> NonNull<libc::c_void> {
         self.ptr
     }
@@ -1271,6 +1318,108 @@ impl Vma {
     pub fn flags(&self) -> ProcmapQueryFlags {
         self.flags
     }
+
+    /// Return VMA name as `XPath`.
+    pub fn name(&self) -> &XPath {
+        XPath::from_bytes(self.name_bytes())
+    }
+
+    /// Return VMA name as bytes.
+    pub fn name_bytes(&self) -> &[u8] {
+        let len = memchr(0, &self.name).unwrap_or(PATH_MAX);
+        &self.name[..len]
+    }
+}
+
+impl TryFrom<SydMemoryMap> for Vma {
+    type Error = Errno;
+
+    fn try_from(map: SydMemoryMap) -> Result<Self, Self::Error> {
+        let map = map.0;
+        let (start, end) = (map.address.0, map.address.1);
+        let ptr = NonNull::new(start as *mut _).ok_or(Errno::EINVAL)?;
+        let len = end
+            .checked_sub(start)
+            .ok_or(Errno::EINVAL)
+            .map(usize::try_from)?
+            .or(Err(Errno::EINVAL))
+            .map(NonZeroUsize::new)?
+            .ok_or(Errno::EINVAL)?;
+
+        let mut name = [0u8; PATH_MAX];
+        match &map.pathname {
+            MMapPath::Path(path) => {
+                let bytes = path.as_os_str().as_bytes();
+                let len = bytes.len().min(PATH_MAX);
+                name[..len].copy_from_slice(&bytes[..len]);
+            }
+            MMapPath::Heap => {
+                name[..7].copy_from_slice(b"[heap]\0");
+            }
+            MMapPath::Stack => {
+                name[..8].copy_from_slice(b"[stack]\0");
+            }
+            MMapPath::TStack(tid) => {
+                use std::io::Write;
+                let _ = write!(&mut name[..], "[stack:{tid}]\0");
+            }
+            MMapPath::Vdso => {
+                name[..7].copy_from_slice(b"[vdso]\0");
+            }
+            MMapPath::Vvar => {
+                name[..7].copy_from_slice(b"[vvar]\0");
+            }
+            MMapPath::Vsyscall => {
+                name[..11].copy_from_slice(b"[vsyscall]\0");
+            }
+            MMapPath::Rollup | MMapPath::Vsys(_) | MMapPath::Other(_) | MMapPath::Anonymous => {}
+        }
+
+        Ok(Self::new(ptr, len, map.perms.into(), name))
+    }
+}
+
+impl TryFrom<(ProcmapQuery, [u8; PATH_MAX])> for Vma {
+    type Error = Errno;
+
+    fn try_from((q, name): (ProcmapQuery, [u8; PATH_MAX])) -> Result<Self, Self::Error> {
+        let ptr = NonNull::new(q.vma_start as *mut _).ok_or(Errno::EINVAL)?;
+        let len = q
+            .vma_end
+            .checked_sub(q.vma_start)
+            .ok_or(Errno::EINVAL)
+            .map(usize::try_from)?
+            .or(Err(Errno::EINVAL))
+            .map(NonZeroUsize::new)?
+            .ok_or(Errno::EINVAL)?;
+        let flags = ProcmapQueryFlags::from_bits_truncate(q.vma_flags);
+        Ok(Self::new(ptr, len, flags, name))
+    }
+}
+
+impl fmt::Display for Vma {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        let start = self.addr();
+        let end = start.saturating_add(self.len().get());
+        let flags = self.flags();
+        let name = self.name();
+        write!(f, "{start:x}-{end:x} {flags} {name}")
+    }
+}
+
+impl Serialize for Vma {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        let start = self.addr();
+        let end = start.saturating_add(self.len().get());
+        let mut map = serializer.serialize_map(Some(3))?;
+        map.serialize_entry("addr", &[start, end])?;
+        map.serialize_entry("perm", &self.flags())?;
+        map.serialize_entry("name", &self.name())?;
+        map.end()
+    }
 }
 
 /// Get the VMA covering `addr` in process `pid`.
@@ -1278,7 +1427,7 @@ impl Vma {
 /// Tries the ioctl(2)-based fast path on Linux>=6.11,
 /// otherwise falls back to parsing _proc_pid_maps_(5).
 ///
-/// On success returns `(vma_start, vma_length, vma_flags)`.
+/// On success returns `Vma`.
 /// If no VMA covers `addr`, returns `Err(Errno::ENOENT)`.
 pub fn proc_get_vma(pid: Pid, addr: u64) -> Result<Vma, Errno> {
     if *HAVE_PROCMAP_QUERY {
@@ -1289,18 +1438,10 @@ pub fn proc_get_vma(pid: Pid, addr: u64) -> Result<Vma, Errno> {
     // Fallback: Sequentially scan /proc/$pid/maps.
     let maps = proc_maps(pid)?;
 
-    for SydMemoryMap(map) in maps {
-        let (start, end) = (map.address.0, map.address.1);
+    for map in maps {
+        let (start, end) = (map.0.address.0, map.0.address.1);
         if (start..end).contains(&addr) {
-            let ptr = NonNull::new(start as *mut _).ok_or(Errno::EINVAL)?;
-            let len = end
-                .checked_sub(start)
-                .ok_or(Errno::EINVAL)
-                .map(usize::try_from)?
-                .or(Err(Errno::EINVAL))
-                .map(NonZeroUsize::new)?
-                .ok_or(Errno::EINVAL)?;
-            return Ok(Vma::new(ptr, len, map.perms.into()));
+            return Vma::try_from(map);
         }
     }
 
@@ -1317,20 +1458,93 @@ fn procmap_query_get_vma(pid: Pid, addr: u64) -> Result<Vma, Errno> {
     // Call PROCMAP_QUERY ioctl(2).
     //
     // No name or build-id buffers needed.
-    let q = procmap_query(&maps, flags, addr, None, None)?;
+    let mut name = [0u8; PATH_MAX];
+    let q = procmap_query(&maps, flags, addr, Some(&mut name), None)?;
 
-    let ptr = NonNull::new(q.vma_start as *mut _).ok_or(Errno::EINVAL)?;
-    let len = q
-        .vma_end
-        .checked_sub(q.vma_start)
-        .ok_or(Errno::EINVAL)
-        .map(usize::try_from)?
-        .or(Err(Errno::EINVAL))
-        .map(NonZeroUsize::new)?
-        .ok_or(Errno::EINVAL)?;
-    let flags = ProcmapQueryFlags::from_bits_truncate(q.vma_flags);
+    Vma::try_from((q, name))
+}
 
-    Ok(Vma::new(ptr, len, flags))
+/// Get all VMAs matching `flags` in process `pid`.
+///
+/// Tries the ioctl(2)-based fast path on Linux>=6.11,
+/// otherwise falls back to parsing _proc_pid_maps_(5).
+///
+/// On success returns `Vec<Vma>`.
+pub fn proc_find_vma(pid: Pid, flags: ProcmapQueryFlags) -> Result<Vec<Vma>, Errno> {
+    if *HAVE_PROCMAP_QUERY {
+        // Use the efficient PROCMAP_QUERY ioctl(2) on Linux>=6.11.
+        return procmap_query_find_vma(pid, flags);
+    }
+
+    // Fallback: Sequentially scan /proc/$pid/maps.
+    let maps = proc_maps(pid)?;
+    let mut vmas = Vec::new();
+
+    for map in maps {
+        let perms = ProcmapQueryFlags::from(map.0.perms);
+
+        if flags.contains(ProcmapQueryFlags::VMA_READABLE)
+            && !perms.contains(ProcmapQueryFlags::VMA_READABLE)
+        {
+            continue;
+        }
+
+        if flags.contains(ProcmapQueryFlags::VMA_WRITABLE)
+            && !perms.contains(ProcmapQueryFlags::VMA_WRITABLE)
+        {
+            continue;
+        }
+
+        if flags.contains(ProcmapQueryFlags::VMA_EXECUTABLE)
+            && !perms.contains(ProcmapQueryFlags::VMA_EXECUTABLE)
+        {
+            continue;
+        }
+
+        if flags.contains(ProcmapQueryFlags::VMA_SHARED)
+            && !perms.contains(ProcmapQueryFlags::VMA_SHARED)
+        {
+            continue;
+        }
+
+        if flags.contains(ProcmapQueryFlags::FILE_BACKED_VMA)
+            && !matches!(map.0.pathname, MMapPath::Path(_))
+        {
+            continue;
+        }
+
+        vmas.push(Vma::try_from(map)?);
+    }
+
+    Ok(vmas)
+}
+
+/// Get all VMAs matching `flags` in process `pid` using PROCMAP_QUERY ioctl(2).
+pub fn procmap_query_find_vma(pid: Pid, flags: ProcmapQueryFlags) -> Result<Vec<Vma>, Errno> {
+    // Open /proc/$pid/maps.
+    let maps = proc_maps_open(pid)?;
+
+    let query_flags = flags | ProcmapQueryFlags::COVERING_OR_NEXT_VMA;
+
+    let mut vmas = Vec::new();
+    let mut addr = 0u64;
+
+    loop {
+        // Call PROCMAP_QUERY ioctl(2).
+        //
+        // No name or build-id buffers needed.
+        let mut name = [0u8; PATH_MAX];
+        match procmap_query(&maps, query_flags, addr, Some(&mut name), None) {
+            Ok(q) => {
+                vmas.push(Vma::try_from((q, name))?);
+                addr = q.vma_end;
+            }
+            Err(Errno::ENOENT) => break,
+            Err(errno) => return Err(errno),
+        }
+    }
+
+    Ok(vmas)
 }
 
 /// Parses /proc/pid/environ and returns a HashMap.
@@ -1349,11 +1563,7 @@ pub fn proc_environ_open(pid: Pid) -> Result<OwnedFd, Errno> {
 
 /// Parses /proc/pid/environ from given Fd and returns a HashMap.
 pub fn proc_environ_read(fd: OwnedFd) -> Result<SydHashMap<OsString, OsString>, Errno> {
-    let mut buf = Vec::new();
-    File::from(fd)
-        .read_to_end(&mut buf)
-        .map_err(|err| err2no(&err))?;
-
+    let buf = read_all(fd)?;
     let mut map = SydHashMap::default();
     for slice in buf.split(|b| *b == 0) {
         // slice will be in the form key=var, so split on the first equals sign.
@@ -1364,7 +1574,6 @@ pub fn proc_environ_read(fd: OwnedFd) -> Result<SydHashMap<OsString, OsString>,
                 OsStr::from_bytes(v).to_os_string(),
             );
         };
-        //let env = OsStr::from_bytes(slice);
     }
 
     Ok(map)
@@ -1531,7 +1740,7 @@ pub fn proc_set_at_secure(pid: Pid, elf_type: ElfType) -> Result<(), Errno> {
             Some(file) => {
                 file.seek(SeekFrom::Start(base_addr as u64))
                     .map_err(|err| err2no(&err))?;
-                read_all(file, buf)
+                read_buf(file, buf)
             }
             None => {
                 let local_len = buf.len();
@@ -1835,7 +2044,7 @@ pub fn proc_max_open_files(pid: Pid) -> Result<LimitValue, Errno> {
     map_result(parse_max_open_files(read_to_end(file, &mut buf)?))
 }
 
-/// Pick a random fd number and check via KCMP_FILE whether it’s open in `pid`.
+/// Pick a random fd number and check via KCMP_FILE whether it's open in `pid`.
 /// On success returns the fd; if none are found after probing the full range,
 /// returns `Err(Errno::EBADF)`. Other errors (permissions, invalid pid, etc.)
 /// are returned immediately.
@@ -2002,7 +2211,7 @@ pub fn proc_kernel_taint() -> Result<KernelTaintFlags, Errno> {
 
     // Read up to 24 bytes + 1 byte for \n.
     let mut data = [0u8; 25];
-    let nread = read_all(fd, &mut data)?;
+    let nread = read_buf(fd, &mut data)?;
 
     // Parse decimal -> u64, then map to flags (retain unknown higher bits).
     let val = btoi::<u64>(data[..nread].trim_ascii()).or(Err(Errno::EINVAL))?;
@@ -2469,4 +2678,24 @@ mod tests {
     fn test_proc_kernel_randomize_va_space() {
         matches!(proc_kernel_randomize_va_space(), Ok(0 | 1 | 2));
     }
+
+    #[test]
+    fn test_proc_find_vma() {
+        if !setup() {
+            return;
+        }
+
+        let pid = Pid::this();
+        for flags in [
+            ProcmapQueryFlags::VMA_READABLE,
+            ProcmapQueryFlags::VMA_EXECUTABLE,
+            ProcmapQueryFlags::VMA_READABLE | ProcmapQueryFlags::VMA_EXECUTABLE,
+        ] {
+            let vmas = proc_find_vma(pid, flags).unwrap();
+            assert!(!vmas.is_empty());
+            for vma in vmas {
+                assert!(vma.flags().contains(flags));
+            }
+        }
+    }
 }
