@@ -1,77 +1,52 @@
 use std::{
-    cmp::Ordering,
     ffi::CString,
-    hint::unreachable_unchecked,
     os::{
-        fd::{AsFd, AsRawFd, BorrowedFd, FromRawFd, OwnedFd},
+        fd::{AsFd, AsRawFd, FromRawFd, IntoRawFd, OwnedFd},
         unix::io::RawFd,
     },
-    ptr,
 };
 
 use libseccomp::ScmpFilterContext;
 use nix::{
     errno::Errno,
     fcntl::OFlag,
-    libc::c_char,
     sys::{
-        ptrace::{cont, seize, Options},
+        ptrace::{cont, Options},
         signal::{kill, Signal},
         wait::{Id, WaitPidFlag},
     },
-    unistd::{isatty, read, setpgid, tcsetpgrp, write, Pid},
+    unistd::{read, setpgid, write, Pid},
 };
 
 use crate::{
-    compat::{waitid, WaitStatus},
-    config::CHLD_STACK_SIZE,
-    fs::pidfd_getfd,
+    compat::{clone3, waitid, CloneArgs, WaitStatus, PTRACE_SEIZE},
+    fd::pidfd_getfd,
+    log::LOG_FD,
     retry::retry_on_eintr,
     rng::duprand,
-    unshare::{child, config::Config, Child, Command, Executable, IoctlMap},
+    sandbox::RawIoctlMap,
+    unshare::{child, config::Config, Child, Command},
 };
 
 type ChildPreExecFunc = Box<dyn Fn() -> Result<(), Errno>>;
 type PipePair = ((OwnedFd, OwnedFd), (OwnedFd, OwnedFd));
 
-#[derive(Debug)]
-pub enum Exe<'a> {
-    Library(&'a libloading::os::unix::Library),
-    Program((*const c_char, Vec<*const c_char>)),
-}
-
-pub struct ChildInfo<'a> {
-    pub exe: Exe<'a>,
+pub struct ChildInfo {
     pub cfg: Config,
+    pub exe_file: CString,
+    pub exe_args: Vec<CString>,
     pub pre_exec: Option<ChildPreExecFunc>,
     pub pty_fd: Option<RawFd>,
-    pub ioctl_denylist: Option<IoctlMap>,
+    pub ioctl_denylist: Option<RawIoctlMap>,
     pub seccomp_filter: Option<ScmpFilterContext>,
     pub seccomp_pipefd: PipePair,
 }
 
-fn raw_with_null(arr: &Vec<CString>) -> Vec<*const c_char> {
-    let mut vec = Vec::with_capacity(arr.len().saturating_add(1));
-    for i in arr {
-        vec.push(i.as_ptr());
-    }
-    vec.push(ptr::null());
-    vec
-}
-
 impl Command {
     /// Spawn the command and return a handle that can be waited for
     pub fn spawn(mut self) -> Result<Child, Errno> {
-        let exe = match self.exe {
-            Executable::Library(ref lib) => Exe::Library(lib),
-            Executable::Program((ref filename, ref args)) => {
-                let c_args = raw_with_null(args);
-                Exe::Program((filename.as_ptr(), c_args))
-            }
-        };
-
-        let mut pid_fd: libc::c_int = -1;
-        let clone_flags = libc::SIGCHLD | libc::CLONE_PIDFD;
+        let exe_file = self.exe_file.take().ok_or(Errno::EFAULT)?;
+        let exe_args = self.exe_args.take().ok_or(Errno::EFAULT)?;
 
         // SAFETY: Child owns its copy of the pipes,
         // and is responsible for closing them.
@@ -89,7 +64,8 @@ impl Command {
         };
 
         let child_info = Box::new(ChildInfo {
-            exe,
+            exe_file,
+            exe_args,
             cfg: self.config,
             pre_exec: std::mem::take(&mut self.pre_exec),
             pty_fd: std::mem::take(&mut self.pty_fd),
@@ -99,66 +75,75 @@ impl Command {
         });
         let child_info_ptr: *mut libc::c_void = Box::into_raw(child_info) as *mut libc::c_void;
 
-        // 2M stack by default, see config.rs.
-        let mut stack = [0u8; CHLD_STACK_SIZE];
-
-        // SAFETY: nix's clone does not support CLONE_PIDFD,
-        // so we use libc::clone instead.
-        let child = unsafe {
-            let ptr = stack.as_mut_ptr().add(stack.len());
-            let ptr_aligned = ptr.sub(ptr as usize % 16);
-            libc::clone(
-                child::child_after_clone as extern "C" fn(*mut libc::c_void) -> libc::c_int,
-                ptr_aligned as *mut libc::c_void,
-                clone_flags,
-                child_info_ptr,
-                &mut pid_fd,
-            )
+        let mut pid_fd: libc::c_int = -1;
+        let mut clone_args = CloneArgs {
+            flags: libc::CLONE_PIDFD as u64,
+            pidfd: (&raw mut pid_fd) as u64,
+            exit_signal: libc::SIGCHLD as u64,
+            ..Default::default()
         };
 
+        let child = match clone3(&mut clone_args) {
+            Ok(child) => child,
+            Err(errno) => {
+                // SAFETY: Reconstruct and drop.
+                drop(unsafe { Box::from_raw(child_info_ptr as *mut ChildInfo) });
+                return Err(errno);
+            }
+        };
+
+        if child.as_raw() == 0 {
+            child::child_after_clone(child_info_ptr);
+        }
+
+        // SAFETY: Put Syd into a new process group for signal safety.
+        // This is best-effort as we may get EPERM here, e.g. under
+        // podman, however the same call in the sandbox process is not
+        // best-effort and fails on error.
+        if self.config.make_group_leader {
+            let _ = setpgid(Pid::from_raw(0), Pid::from_raw(0));
+        }
+
+        // SAFETY: Reconstruct and forget pipe fds,
+        // after_start is going to handle them.
+        let child_info = unsafe { Box::from_raw(child_info_ptr as *mut ChildInfo) };
+        let _ = child_info.seccomp_pipefd.0 .0.into_raw_fd();
+        let _ = child_info.seccomp_pipefd.0 .1.into_raw_fd();
+        let _ = child_info.seccomp_pipefd.1 .0.into_raw_fd();
+        let _ = child_info.seccomp_pipefd.1 .1.into_raw_fd();
+
         // SAFETY: Randomize the pid FD for hardening.
-        // The created fd is O_CLOEXEC too, and we'll
-        // send the number to the child to close it.
         // O_EXCL closes oldfd on success.
         let pid_fd = duprand(pid_fd, OFlag::O_CLOEXEC | OFlag::O_EXCL)?;
 
-        // SAFETY: duprand returns a valid FD on success.
-        let pid_fd = unsafe { BorrowedFd::borrow_raw(pid_fd) };
+        // SAFETY: Randomize the log FD for hardening.
+        // O_EXCL closes oldfd on success.
+        let log_fd = LOG_FD.load(std::sync::atomic::Ordering::Relaxed);
+        if log_fd >= 0 {
+            let log_fd = duprand(log_fd, OFlag::O_CLOEXEC | OFlag::O_EXCL)?;
+            LOG_FD.store(log_fd.into_raw_fd(), std::sync::atomic::Ordering::Relaxed);
+        } // else logging is disabled.
 
-        match child.cmp(&0) {
-            Ordering::Less => {
-                // SAFETY: Reconstruct and drop.
-                let _ = unsafe { Box::from_raw(child_info_ptr as *mut ChildInfo) };
-                Err(Errno::last())
-            }
-            Ordering::Greater => {
-                let child = Pid::from_raw(child);
+        let seccomp_fd = match self.after_start(child, &pid_fd) {
+            Ok(seccomp_fd) => seccomp_fd,
+            Err(e) => loop {
+                match waitid(Id::PIDFd(pid_fd.as_fd()), WaitPidFlag::WEXITED) {
+                    Ok(WaitStatus::Exited(_, errno)) => return Err(Errno::from_raw(errno)),
+                    Err(Errno::EINTR) => {}
+                    _ => return Err(e),
+                }
+            },
+        };
 
-                let seccomp_fd = match self.after_start(child, pid_fd) {
-                    Ok(seccomp_fd) => seccomp_fd,
-                    Err(e) => loop {
-                        match waitid(Id::PIDFd(pid_fd.as_fd()), WaitPidFlag::WEXITED) {
-                            Ok(WaitStatus::Exited(_, errno)) => return Err(Errno::from_raw(errno)),
-                            Err(Errno::EINTR) => {}
-                            _ => return Err(e),
-                        }
-                    },
-                };
-
-                Ok(Child {
-                    pid: child.into(),
-                    pid_fd: pid_fd.as_raw_fd(),
-                    seccomp_fd,
-                    status: None,
-                })
-            }
-            // SAFETY: This can never happen because clone child
-            // jumps to the specified function.
-            _ => unsafe { unreachable_unchecked() },
-        }
+        Ok(Child {
+            pid: child.into(),
+            pid_fd: pid_fd.into_raw_fd(),
+            seccomp_fd: seccomp_fd.into_raw_fd(),
+            status: None,
+        })
     }
 
-    fn after_start<Fd: AsFd>(mut self, pid: Pid, pid_fd: Fd) -> Result<RawFd, Errno> {
+    fn after_start<Fd: AsFd>(mut self, pid: Pid, pid_fd: Fd) -> Result<OwnedFd, Errno> {
         if self.config.stop {
             // Seize the process for tracing.
             // This must happen before reading the seccomp fd.
@@ -192,7 +177,16 @@ impl Command {
             //    The typical error case is EPERM which means parent is
             //    strace or YAMA is active.
             assert_eq!(
-                seize(pid, ptrace_options),
+                // SAFETY: In libc, we trust.
+                Errno::result(unsafe {
+                    libc::ptrace(
+                        PTRACE_SEIZE,
+                        pid.as_raw(),
+                        0,
+                        ptrace_options.bits() as *mut libc::c_void,
+                    )
+                })
+                .map(drop),
                 Ok(()),
                 "YAMA or strace? Use with strace -f syd -pD ..."
             );
@@ -205,7 +199,9 @@ impl Command {
                 WaitStatus::PtraceEvent(pid, libc::SIGSTOP, libc::PTRACE_EVENT_STOP)
             );
             // SAFETY: nix does not have a wrapper for PTRACE_LISTEN.
-            Errno::result(unsafe { libc::ptrace(libc::PTRACE_LISTEN, pid.as_raw(), 0, 0) })?;
+            Errno::result(unsafe {
+                libc::ptrace(crate::compat::PTRACE_LISTEN, pid.as_raw(), 0, 0)
+            })?;
             // Step 3: Successfully attached, resume the process.
             // We have to do a simple signal ping-pong here but
             // it's done once and it's worth the trouble.
@@ -227,24 +223,6 @@ impl Command {
             cont(pid, Some(Signal::SIGCONT))?;
         }
 
-        let stdin_isatty = isatty(std::io::stdin()).unwrap_or(false);
-        if self.config.make_group_leader {
-            // SAFETY: Put sandbox process in a new process group
-            // to isolate Syd processes from the sandbox processes.
-            setpgid(pid, pid)?;
-
-            // Check if standard input is a TTY.
-            if stdin_isatty {
-                // Set foreground process group to the sandbox process.
-                tcsetpgrp(std::io::stdin(), pid)?;
-            }
-        }
-
-        // SAFETY: Note we don't want to put Syd into a new process
-        // group here so as to defend the parent process against
-        // signals in cases when Landlock protection is not available.
-        // Landlock signal scopes are new in ABI 6, new in Linux>=6.12.
-
         if let Some(ref mut callback) = self.before_unfreeze {
             #[expect(clippy::cast_sign_loss)]
             callback(i32::from(pid) as u32)?;
