@@ -32,7 +32,8 @@ use crate::{
     cookie::{OPENAT2_COOKIE_ARG4, OPENAT2_COOKIE_ARG5},
     err::{err2no, scmp2no, SydJoinHandle, SydResult},
     error,
-    fs::{closeexcept, seccomp_notify_id_valid, tgkill},
+    fd::closeexcept,
+    fs::{seccomp_notify_id_valid, tgkill},
     info,
     proc::{proc_status_open, proc_status_read},
     retry::retry_on_eintr,
@@ -81,7 +82,7 @@ pub(crate) struct Interrupter {
     transit_gids: Vec<(Gid, Gid)>,
 
     should_exit: Arc<AtomicBool>,
-    cache: Arc<WorkerCache<'static>>,
+    cache: Arc<WorkerCache>,
 }
 
 impl Interrupter {
@@ -91,7 +92,7 @@ impl Interrupter {
         transit_uids: &[(Uid, Uid)],
         transit_gids: &[(Gid, Gid)],
         should_exit: Arc<AtomicBool>,
-        cache: Arc<WorkerCache<'static>>,
+        cache: Arc<WorkerCache>,
     ) -> Self {
         Self {
             flags,
@@ -179,7 +180,7 @@ impl Interrupter {
                         Ok(ctx) => ctx,
                         Err(error) => {
                             let errno = error.errno().unwrap_or(Errno::ENOSYS);
-                            alert!("ctx": "boot", "op": "confine_interrupt_thread",
+                            alert!("ctx": "boot", "op": "confine_int_thread",
                                 "msg": format!("failed to confine: {error}"),
                                 "err": errno as i32);
                             std::process::exit(101);
@@ -191,7 +192,7 @@ impl Interrupter {
                     // because this confinement is a critical safety feature.
                     if let Err(error) = ctx.load() {
                         let errno = scmp2no(&error).unwrap_or(Errno::ENOSYS);
-                        alert!("ctx": "boot", "op": "confine_interrupt_thread",
+                        alert!("ctx": "boot", "op": "confine_int_thread",
                             "msg": format!("failed to confine: {error}"),
                             "err": errno as i32);
                         std::process::exit(101);
@@ -200,7 +201,7 @@ impl Interrupter {
                     let safe_setid = self
                         .flags
                         .intersects(Flags::FL_ALLOW_SAFE_SETUID | Flags::FL_ALLOW_SAFE_SETGID);
-                    info!("ctx": "confine", "op": "confine_interrupt_thread",
+                    info!("ctx": "confine", "op": "confine_int_thread",
                         "msg": format!("interrupt thread confined with{} SROP mitigation",
                             if safe_setid { "out" } else { "" }));
                 } else {
@@ -217,7 +218,7 @@ impl Interrupter {
     fn main(self, tgid: Pid) -> SydResult<()> {
         loop {
             // Wait for one cycle.
-            std::thread::sleep(INT_CYCLE_TIME);
+            std::thread::sleep(INT_CYCLE_TIME.into());
 
             // Unblock invalidated blocking system calls.
             {
